@@ -1,4 +1,4 @@
-/*  $Id: bags.pl,v 1.2 1996/01/22 15:16:06 jan Exp $
+/*  $Id: bags.pl,v 1.3 1997/05/06 09:59:50 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     jan@swi.psy.uva.nl
@@ -51,7 +51,7 @@ bagof(Gen, Goal, Bag) :-
 	Bag \== [].
 
 assert_bag(Templ, G) :-
-	$record_bag(-(mark, -)), 
+	$record_bag(-), 
 	G,
 	    $record_bag(Templ), 
 	fail.
