@@ -1,4 +1,4 @@
-/*  $Id: syspred.pl,v 1.34 1997/08/07 07:57:26 jan Exp $
+/*  $Id: syspred.pl,v 1.35 1998/08/06 13:25:22 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     jan@swi.psy.uva.nl
@@ -42,6 +42,8 @@
 	, shell/2
 	, shell/1
 	, shell/0
+	, on_signal/3
+	, current_signal/3
 	, open_shared_object/2
 	, open_shared_object/3
 	, format/1
@@ -609,6 +611,26 @@ shell :-
 shell :-
 	shell('/bin/sh').
 
+		 /*******************************
+		 *	      SIGNALS		*
+		 *******************************/
+
+:- module_transparent
+	on_signal/3.
+
+on_signal(Signal, Old, New) :-
+	atom(Signal), !,
+	$on_signal(_Num, Signal, Old, New).
+on_signal(Signal, Old, New) :-
+	integer(Signal), !,
+	$on_signal(Signal, _Name, Old, New).
+on_signal(Signal, _Old, _New) :-
+	throw(error(type_error(signal, Signal), on_signal/3)).
+
+current_signal(Name, Id, Handler) :-
+	between(1, 32, Id),
+	$on_signal(Name, Id, Handler, Handler).
+
 
 		 /*******************************
 		 *	      DLOPEN		*
