@@ -1,4 +1,4 @@
-/*  $Id: toplevel.pl,v 1.34 1997/11/04 10:38:22 jan Exp $
+/*  $Id: toplevel.pl,v 1.36 1998/08/06 13:25:24 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     jan@swi.psy.uva.nl
@@ -35,7 +35,7 @@ $welcome :-
 	Patch is Version mod 100,
 	$ttyformat('Welcome to SWI-Prolog (Version ~w.~w.~w)~n',
 		   [Major, Minor, Patch]),
-	$ttyformat('Copyright (c) 1993-1997 University of Amsterdam.  '),
+	$ttyformat('Copyright (c) 1993-1998 University of Amsterdam.  '),
 	$ttyformat('All rights reserved.~n~n'),
 	$ttyformat('For help, use ?- help(Topic). or ?- apropos(Word).~n~n').
 
@@ -248,7 +248,7 @@ $toplevel :-
 	$ttyformat('[halt]~n', []).		
 
 $runtoplevel :-
-	$option(top_level, TopLevelAtom, TopLevelAtom), 
+	$option(toplevel, TopLevelAtom, TopLevelAtom), 
 	term_to_atom(TopLevel, TopLevelAtom), 
 	user:TopLevel.
 
@@ -479,13 +479,16 @@ answer_respons(Char, redo) :-
 	trace,
 	$format_if_tty('; [trace]~n').
 answer_respons(Char, continue) :-
-	memberchk(Char, [0'c, 0' , 10, 13, 0'y, 0'Y]), !.
+	memberchk(Char, [0'c, 0'a, 0' , 10, 13, 0'y, 0'Y]), !.
 answer_respons(0'b, show_again) :- !,
 	break.
 answer_respons(Char, show_again) :-
 	print_predicate(Char, Pred), !,
 	$format_if_tty('~w~n', [Pred]),
 	flag($toplevel_print_predicate, _, Pred).
+answer_respons(-1, show_again) :- !,
+	$format_if_tty('EOF: exit~n'),
+	halt(0).
 answer_respons(_, again) :-
 	$ttyformat('~nUnknown action (h for help)~nAction? '),
 	ttyflush.
@@ -496,10 +499,10 @@ print_predicate(0'p, print).
 
 show_toplevel_usage :-
 	$ttyformat('~nActions:~n'),
-	$ttyformat('; (n, r):     redo    t:               trace & redo~n'),
-	$ttyformat('b:            break   c (ret, space):  continue~n'),
-	$ttyformat('d:            display p                print~n'),
-	$ttyformat('w:            write   h (?):           help~n').
+	$ttyformat('; (n, r):     redo    t:                 trace & redo~n'),
+	$ttyformat('b:            break   c (a, RET, space): continue~n'),
+	$ttyformat('d:            display p                  print~n'),
+	$ttyformat('w:            write   h (?):             help~n').
 
 $format_if_tty(Fmt) :-
 	$format_if_tty(Fmt, []).
@@ -532,14 +535,6 @@ $time_call(Goal, yes) :-
 	Goal, !.
 $time_call(_Goal, no).
 
-unhandled_exception(false, Term) :- !,
-        $warning('Unhandled exception'),
-	print_message(error, Term),
-	$ttyformat('~nNo~n').
-unhandled_exception(true, _Term) :-
-	$warning('Unhandled exception'),
-	$ttyformat('~nNo~n').
-
 
 		 /*******************************
 		 *	    EXPANSION		*
