@@ -1,4 +1,4 @@
-/*  $Id: latex2html.pl,v 1.8 1997/07/11 08:29:26 jan Exp $
+/*  $Id: latex2html.pl,v 1.11 1998/07/06 12:42:31 jan Exp $
 
     Designed and implemented by Jan Wielemaker
     E-mail: jan@swi.psy.uva.nl
@@ -385,6 +385,7 @@ language_map(table,	'Table').
 #(em(Text),		[html('<EM>'),     Text, html('</EM>')]).
 #(b(Text),		[html('<B>'),      Text, html('</B>')]).
 #(i(Text),		[html('<I>'),      Text, html('</I>')]).
+#(tt(Text),		[html('<TT>'),	   Text, html('</TT>')]).
 #(center(Text),		[html('<CENTER>'), Text, html('</CENTER>')]).
 #(right(Text),		[html('<RIGHT>'),  Text, html('</RIGHT>')]).
 #(quote(Text),		[html('<BLOCKQUOTE>'), Text, html('</BLOCKQUOTE>')]).
@@ -799,6 +800,9 @@ cmd(footheight(_), []).
 cmd(vfill, [html('<P>')]).
 cmd(vfil, [html('<P>')]).
 cmd(hfill, []).
+cmd(/, []).
+cmd(and, [html('<BR>')]).
+cmd(leavevmode, []).
 cmd(parskip(_), []).
 cmd(parindent(_), []).
 cmd(raggedright, []).			% Always in HTML
@@ -894,9 +898,12 @@ cmd(bibliography({_}), HTML) :-			% \bibliography
 cmd(bibliographystyle(_), []).			% \bibliographystyle
 cmd(',', []).					% \,
 
+cmd(emph({Tex}), #em(+Tex)).			% \emph{text}
+
 cmd(year,	Year) :-			% \year
 	get_time(Time),
 	convert_time(Time, Year, _, _, _, _, _, _).
+cmd('LaTeX',	'LaTeX').			% \LaTeX
 
 cmd(index({Term}), #label(RefName, [])) :-	% \index
 	translate_index(Term, RefName).
@@ -1009,7 +1016,11 @@ cmd(psfig({Spec}), html(Img)) :-
 	(   keep_figures(true),
 	    exists_file(OutFile)
 	->  true
-	;   ps2gif(tex(Base), OutFile)
+	;   (   is_absolute_file_name(Base)
+	    ->	FileSpec = Base
+	    ;	FileSpec = tex(Base)
+	    ),
+	    ps2gif(FileSpec, OutFile)
 	).
 cmd(postscript({_Width}, {File}, Title),
     [ LabelHTML,
