@@ -1,4 +1,4 @@
-/*  $Id: progman.pl,v 1.9 1996/10/30 17:41:49 jan Exp $
+/*  $Id: progman.pl,v 1.10 1997/04/04 09:05:42 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     jan@swi.psy.uva.nl
@@ -121,7 +121,7 @@ progman_setup :-
 	prolog_to_os_filename(PlExe, OsPlExe),
 
 	progman_make_group(Group),
-	progman_make_item(Group, 'SWI-Prolog', OsPlExe, 'c:').
+	progman_make_item(Group, 'SWI-Prolog', OsPlExe, 'c:'),
 	explain(end).
 
 
