@@ -94,3 +94,11 @@
 /* Define max size of mmapp()ed stacks.  See test/mmap.c */
 #undef MMAP_STACKSIZE
 
+/* Define if the type rlim_t is defined by <sys/resource.h> */
+#undef HAVE_RLIM_T
+
+/* Define to 1 if &&label and goto *var is supported (GCC-2) */
+#undef O_LABEL_ADDRESSES
+
+/* Define to 1 not to use SIGSEGV for guarding stack-overflows */
+#undef NO_SEGV_HANDLING 
