@@ -1,4 +1,4 @@
-/*  $Id: pl-ext.c,v 1.61 1997/10/28 13:40:52 jan Exp $
+/*  $Id: pl-ext.c,v 1.65 1998/08/06 13:26:55 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -36,12 +36,12 @@ Link all foreign language predicates.  The arguments to FRG are:
 
 Flags almost always is TRACE_ME.  Additional common flags:
 
-	TRANSPARENT		Predicate is module transparent
+	METAPRED		Predicate is module transparent
 	NONDETERMINISTIC	Predicate can be resatisfied
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define NDET NONDETERMINISTIC		/* make a bit shorter */
-#define META TRANSPARENT		/* same */
+#define NDET NONDETERMINISTIC	/* make a bit shorter */
+#define META METAPRED		/* same */
 
 #define FRG(n, a, f, flags) { n, f, flags, a }
 
@@ -90,6 +90,7 @@ static const struct foreign {
   FRG("file_directory_name",	2, pl_file_dir_name,		TRACE_ME),
   FRG("file_name_extension",	3, pl_file_name_extension,	TRACE_ME),
   FRG("prolog_to_os_filename",	2, pl_prolog_to_os_filename,	TRACE_ME),
+  FRG("$mark_executable",	1, pl_mark_executable,		TRACE_ME),
 #ifdef __WIN32__
   FRG("win_exec",		2, pl_win_exec,			TRACE_ME),
 #ifdef O_XOS
@@ -97,6 +98,8 @@ static const struct foreign {
 #endif
 #endif
 
+  FRG("$on_signal",		4, pl_on_signal,	   META|TRACE_ME),
+
   FRG("fileerrors",		2, pl_fileerrors,		TRACE_ME),
   FRG("$syntaxerrors",		2, pl_syntaxerrors,		TRACE_ME),
   FRG("chdir",			1, pl_chdir,			TRACE_ME),
@@ -109,7 +112,7 @@ static const struct foreign {
   FRG("wildcard_match",		2, pl_wildcard_match,		TRACE_ME),
   FRG("$apropos_match",		2, pl_apropos_match,		TRACE_ME),
   FRG("$argv",			1, pl_argv,			TRACE_ME),
-  FRG("$option",		3, pl_option,			TRACE_ME),
+  FRG("$option",		3, pl_option,		   NDET|TRACE_ME),
   FRG("convert_time",		8, pl_convert_time,		TRACE_ME),
   FRG("sleep",			1, pl_sleep,			TRACE_ME),
   FRG("break",			0, pl_break,			TRACE_ME),
@@ -166,12 +169,20 @@ static const struct foreign {
   FRG("free_variables",		2, pl_free_variables,		TRACE_ME),
   FRG("$e_free_variables",	2, pl_e_free_variables,		TRACE_ME),
 
-  FRG("$open_wic",		2, pl_open_wic,			TRACE_ME),
+  FRG("$open_wic",		1, pl_open_wic,			TRACE_ME),
   FRG("$close_wic",		0, pl_close_wic,		TRACE_ME),
   FRG("$add_directive_wic",	1, pl_add_directive_wic,	TRACE_ME),
   FRG("$import_wic",		2, pl_import_wic,		TRACE_ME),
 
-  FRG("$qlf_put_states",	0, pl_qlf_put_states,		TRACE_ME),
+  FRG("$rc_handle",		1, pl_rc_handle,		TRACE_ME),
+  FRG("$rc_members",		2, pl_rc_members,		TRACE_ME),
+  FRG("$rc_open",		5, pl_rc_open,			TRACE_ME),
+  FRG("$rc_open_archive",	2, pl_rc_open_archive,		TRACE_ME),
+  FRG("$rc_close_archive",	1, pl_rc_close_archive,		TRACE_ME),
+  FRG("$rc_save_archive",	2, pl_rc_save_archive,		TRACE_ME),
+  FRG("$rc_append_file",	5, pl_rc_append_file,		TRACE_ME),
+  FRG("$copy_stream",		2, pl_copy_stream,		TRACE_ME),
+
   FRG("$qlf_start_module",	1, pl_qlf_start_module,		TRACE_ME),
   FRG("$qlf_start_sub_module",	1, pl_qlf_start_sub_module,	TRACE_ME),
   FRG("$qlf_start_file",	1, pl_qlf_start_file,		TRACE_ME),
@@ -179,7 +190,7 @@ static const struct foreign {
   FRG("$qlf_open",		1, pl_qlf_open,			TRACE_ME),
   FRG("$qlf_close",		0, pl_qlf_close,		TRACE_ME),
   FRG("$qlf_load",		2, pl_qlf_load,		   META|TRACE_ME),
-  FRG("$qlf_assert_clause",	1, pl_qlf_assert_clause,	TRACE_ME),
+  FRG("$qlf_assert_clause",	2, pl_qlf_assert_clause,	TRACE_ME),
   FRG("$qlf_info",		4, pl_qlf_info,			TRACE_ME),
 
   FRG("abolish",    		1, pl_abolish1,		   META|TRACE_ME),
@@ -204,6 +215,7 @@ static const struct foreign {
   FRG("recordz",		3, pl_recordz,			TRACE_ME),
   FRG("recorded",		3, pl_recorded,		   NDET|TRACE_ME),
   FRG("erase",			1, pl_erase,			TRACE_ME),
+  FRG("$term_complexity",	3, pl_term_complexity,		TRACE_ME),
   FRG("assert",			1, pl_assertz,		   META|TRACE_ME),
   FRG("asserta",		1, pl_asserta,		   META|TRACE_ME),
   FRG("assertz",		1, pl_assertz,		   META|TRACE_ME),
@@ -292,7 +304,6 @@ static const struct foreign {
 
   FRG("prolog_current_frame",	1, pl_prolog_current_frame,	TRACE_ME),
   FRG("prolog_frame_attribute",	3, pl_prolog_frame_attribute,	TRACE_ME),
-  FRG("$trace_continuation",	1, pl_trace_continuation,	TRACE_ME),
   FRG("prolog_skip_level",	2, pl_skip_level,		0),
 
   FRG("$write_on_atom",		2, pl_write_on_atom,		TRACE_ME),
@@ -374,7 +385,7 @@ static const struct foreign {
   FRG("$check_definition",	1, pl_check_definition,    META|TRACE_ME),
 
 #if O_COMPILE_OR
-  FRG("$alt",			1, pl_alt,			NDET),
+  FRG("$alt",			0, pl_alt,			NDET),
 #endif /* O_COMPILE_OR */
   FRG("$atom_hashstat",		2, pl_atom_hashstat,		TRACE_ME),
   FRG("$tty",			0, pl_tty,			TRACE_ME),
@@ -396,10 +407,12 @@ static const struct foreign {
   FRG("current_stream",		3, pl_current_stream,	   NDET|TRACE_ME),
   FRG("flush_output",		1, pl_flush_output,		TRACE_ME),
   FRG("stream_position",	3, pl_stream_position,		TRACE_ME),
+  FRG("seek",			4, pl_seek,			TRACE_ME),
   FRG("set_input",		1, pl_set_input,		TRACE_ME),
   FRG("set_output",		1, pl_set_output,		TRACE_ME),
   FRG("current_input",		1, pl_current_input,		TRACE_ME),
   FRG("current_output",		1, pl_current_output,		TRACE_ME),
+  FRG("dup_stream",		2, pl_dup_stream,		TRACE_ME),
   FRG("character_count",	2, pl_character_count,		TRACE_ME),
   FRG("line_count",		2, pl_line_count,		TRACE_ME),
   FRG("line_position",		2, pl_line_position,		TRACE_ME),
@@ -429,6 +442,7 @@ static const struct foreign {
   FRG("wait_for_input",		3, pl_wait_for_input,		TRACE_ME),
   FRG("get_time",		1, pl_get_time,			TRACE_ME),
   FRG("size_file",		2, pl_size_file,		TRACE_ME),
+  FRG("$size_stream",		2, pl_size_stream,		TRACE_ME),
   FRG("$default_module",	3, pl_default_module,	   META|TRACE_ME),
 #if O_PROLOG_FUNCTIONS
   FRG("$arithmetic_function",   1, pl_arithmetic_function, META|TRACE_ME),
@@ -469,7 +483,7 @@ bindExtensions(PL_extension *e)
   { short flags = TRACE_ME;
 
     if ( e->flags & PL_FA_NOTRACE )	     flags &= ~TRACE_ME;
-    if ( e->flags & PL_FA_TRANSPARENT )	     flags |= TRANSPARENT;
+    if ( e->flags & PL_FA_TRANSPARENT )	     flags |= METAPRED;
     if ( e->flags & PL_FA_NONDETERMINISTIC ) flags |= NONDETERMINISTIC;
 
     def = lookupProcedure(lookupFunctorDef(lookupAtom(e->predicate_name),
@@ -507,6 +521,7 @@ initBuildIns(void)
 { const struct foreign *f;
   Definition def;
   ExtensionCell ecell;
+  Module m = MODULE_system;
 
   for(f = &foreigns[0]; f->name; f++)
   { functor_t fdef = lookupFunctorDef(lookupAtom(f->name), f->arity);
@@ -523,15 +538,13 @@ initBuildIns(void)
       set(valueFunctor(fdef), INLINE_F);
   }
 
-  PROCEDURE_alt1	     = lookupProcedure(FUNCTOR_alt1, MODULE_system);
-  PROCEDURE_garbage_collect0 = lookupProcedure(FUNCTOR_dgarbage_collect1,
-					       MODULE_system);
-  PROCEDURE_block3	     = lookupProcedure(FUNCTOR_block3, MODULE_system);
-  PROCEDURE_catch3           = lookupProcedure(FUNCTOR_catch3, MODULE_system);
-  PROCEDURE_true0            = lookupProcedure(FUNCTOR_true0, MODULE_system);
-  PROCEDURE_fail0            = lookupProcedure(FUNCTOR_fail0, MODULE_system);
-  PROCEDURE_print_message2   = lookupProcedure(FUNCTOR_print_message2,
-					       MODULE_system);
+  PROCEDURE_alt0	     = lookupProcedure(FUNCTOR_alt0,		  m);
+  PROCEDURE_garbage_collect0 = lookupProcedure(FUNCTOR_dgarbage_collect1, m);
+  PROCEDURE_block3	     = lookupProcedure(FUNCTOR_block3, 		  m);
+  PROCEDURE_catch3           = lookupProcedure(FUNCTOR_catch3, 		  m);
+  PROCEDURE_true0            = lookupProcedure(FUNCTOR_true0, 		  m);
+  PROCEDURE_fail0            = lookupProcedure(FUNCTOR_fail0, 		  m);
+  PROCEDURE_print_message2   = lookupProcedure(FUNCTOR_print_message2, 	  m);
 
   for( ecell = ext_head; ecell; ecell = ecell->next )
     bindExtensions(ecell->extensions);
