@@ -20,7 +20,6 @@ handling times must be cleaned, but that not only holds for this module.
 
 #if defined(__WINDOWS__) || defined(__NT__)
 #include "windows.h"
-#undef TRANSPARENT
 #undef FD_SET
 #undef FD_ISSET
 #undef FD_ZERO
@@ -60,8 +59,8 @@ static struct plfile
 { atom_t	name;			/* name of file */
   atom_t	stream_name;		/* stream identifier name */
   IOSTREAM *	stream;			/* IOSTREAM package descriptor */
-  int		status;			/* F_CLOSED, F_READ, F_WRITE */
-  int		type;			/* ST_FILE, ST_PIPE, ST_STRING */
+  char		status;			/* F_CLOSED, F_READ, F_WRITE */
+  char		type;			/* ST_FILE, ST_PIPE, ST_STRING */
 } *fileTable = (PlFile) NULL;		/* Our file table */
 
 int 	Input;				/* current input */
@@ -122,9 +121,16 @@ pipeHandler(int sig)
 static void
 brokenPipe(int n, atom_t rw)
 { term_t stream = PL_new_term_ref();
+  IOSTREAM *s;
+
+  if ( (s = fileTable[n].stream) )
+    s->bufp = s->buffer;		/* avoid flush on close! */
+
   unifyStreamNo(stream, n);
+
   if ( rw == ATOM_write && n == Output )
     Output = 1;
+
   PL_error(NULL, 0, "Broken pipe", ERR_STREAM_OP, rw, stream);
 }
 
@@ -171,31 +177,31 @@ without the Unix assumptions?
 
     switch(n)
     { case 0:
-	f->name	       = ATOM_user;
-	f->stream_name = ATOM_user_input;
-	f->stream      = Sinput;
-	f->status      = F_READ;
-	f->type	       = ST_TERMINAL;
+	f->name	          = ATOM_user;
+	f->stream_name    = ATOM_user_input;
+	f->stream         = Sinput;
+	f->status         = F_READ;
+	f->type	          = ST_TERMINAL;
 	break;
       case 1:
-	f->name        = ATOM_user;
-	f->stream_name = ATOM_user_output;
-	f->stream      = Soutput;
-	f->status      = F_WRITE;
-	f->type	       = ST_TERMINAL;
+	f->name           = ATOM_user;
+	f->stream_name    = ATOM_user_output;
+	f->stream         = Soutput;
+	f->status         = F_WRITE;
+	f->type	          = ST_TERMINAL;
 	break;
       case 2:
-	f->name        = ATOM_stderr;
-	f->stream_name = ATOM_user_error;
-	f->stream      = Serror;
-	f->status      = F_WRITE;
-	f->type	       = ST_TERMINAL;
+	f->name           = ATOM_stderr;
+	f->stream_name    = ATOM_user_error;
+	f->stream         = Serror;
+	f->status         = F_WRITE;
+	f->type	          = ST_TERMINAL;
 	break;
       default:
-	f->name        = NULL_ATOM;
-        f->stream      = NULL_ATOM;
-	f->type        = ST_FILE;
-	f->status      = F_CLOSED;
+	f->name           = NULL_ATOM;
+        f->stream         = NULL_ATOM;
+	f->type           = ST_FILE;
+	f->status         = F_CLOSED;
     }
   }
 
@@ -219,7 +225,7 @@ void
 dieIO()
 { if ( GD->io_initialised )
   { pl_noprotocol();
-    closeFiles();
+    closeFiles(TRUE);
     PopTty(&ttytab);
   }
 }
@@ -255,16 +261,23 @@ closeStream(int n)
 
 
 void
-closeFiles(void)
-{ int n;
+closeFiles(int all)
+{ volatile int n;
 #if O_PCE
   extern int read_nesting;
   read_nesting = 0;
 #endif
 
   for(n=0; n<maxfiles; n++)
-  { if ( n != protocolStream )
-      closeStream(n);
+  { IOSTREAM *s;
+
+    if ( (s=fileTable[n].stream) )
+    { if ( all || !(s->flags & SIO_NOCLOSE) )
+	closeStream(n);
+      else if ( fileTable[n].status == F_WRITE )
+      {	TRYPIPE(n, ATOM_write, Sflush(s), (void)0);
+      }
+    }
   }
 
   Input = 0;
@@ -394,6 +407,9 @@ getSingleChar(void)
       c = Get0();
   }
 
+  if ( c == 4 || c == 26 )		/* should ask the terminal! */
+    c = -1;
+
   PopTty(&buf);
   debugstatus.suspendTrace--;
   Input = OldIn;
@@ -516,7 +532,8 @@ PutOpenToken(int c)
 
   if ( s->lastc != EOF &&
        ((isAlpha(s->lastc) && isAlpha(c)) ||
-	(isSymbol(s->lastc) && isSymbol(c))) )
+	(isSymbol(s->lastc) && isSymbol(c)) ||
+	c == '(') )
     return Put(' ');
 
   succeed;
@@ -588,6 +605,26 @@ PL_current_output()
 }
 
 
+word
+pl_dup_stream(term_t from, term_t to)
+{ int fn, tn;
+  PlFile f, t;
+
+  if ( (fn = streamNo(from, F_ANY)) < 0 ||
+       (tn = streamNo(to, F_ANY)) < 0 )
+    fail;
+
+  f = &fileTable[fn];
+  t = &fileTable[tn];
+
+  t->stream = f->stream;
+  t->status = f->status;
+  t->type   = f->type;
+
+  succeed;
+}
+
+
 bool
 PL_open_stream(term_t handle, IOSTREAM *s)
 { int n;
@@ -603,7 +640,7 @@ PL_open_stream(term_t handle, IOSTREAM *s)
       else
 	f->status = F_WRITE;
 
-      return PL_unify_integer(handle, n);
+      return setUnifyStreamNo(handle, n);
     }
   }
 
@@ -729,10 +766,10 @@ openStream(term_t file, int mode, int flags)
   if ( n >= maxfiles )			/* non-ISO */
     return PL_error(NULL, 0, NULL, ERR_REPRESENTATION, ATOM_max_files);
 
-  fileTable[n].name = name;
+  fileTable[n].name        = name;
   fileTable[n].stream_name = NULL_ATOM;
-  fileTable[n].type = type;
-  fileTable[n].stream = stream;
+  fileTable[n].type        = type;
+  fileTable[n].stream      = stream;
 
   switch(mode)
   { case F_READ:
@@ -762,7 +799,7 @@ unifyStreamName(term_t f, int n)
 #ifdef HAVE_POPEN
   if ( fileTable[n].type == ST_PIPE )
   { return PL_unify_term(f,
-			 FUNCTOR_pipe1,
+			 PL_FUNCTOR, FUNCTOR_pipe1,
 			   PL_ATOM, fileTable[n].name);
   }
 #endif /*HAVE_POPEN*/
@@ -798,7 +835,8 @@ unifyStreamNo(term_t stream, int n)
     default:
       if ( fileTable[n].stream_name )
 	name = fileTable[n].stream_name;
-      return PL_unify_integer(stream, n);
+      else
+	return PL_unify_integer(stream, n);
   }
 
   return PL_unify_atom(stream, name);
@@ -819,8 +857,12 @@ pl_told()
 
 word
 pl_flush()
-{ if ( fileTable[Output].stream )
-    Sflush(fileTable[Output].stream);
+{ IOSTREAM *s;
+
+  if ( fileTable[Output].status == F_WRITE &&
+       (s=fileTable[Output].stream) )
+  { TRYPIPE(Output, ATOM_write, Sflush(s), fail);
+  }
 
   succeed;
 }
@@ -851,7 +893,10 @@ openProtocol(term_t f, bool appnd)
   pl_noprotocol();
 
   if ( openStream(f, appnd ? F_APPEND : F_WRITE, OPEN_TEXT|OPEN_OPEN) )
-  { protocolStream = Output;
+  { IOSTREAM *s = fileTable[Output].stream;
+
+    s->flags |= SIO_NOCLOSE;
+    protocolStream = Output;
     Output = out;
 
     succeed;
@@ -1081,17 +1126,12 @@ pl_put(term_t c)
   char *s;
 
   if ( PL_get_integer(c, &chr) )
-  { if (chr < 0 || chr > 255)
-      goto err;
-    Put(chr);
+  { if ( chr >= 0 && chr <= 255 )
+      return Put(chr);
   } else if ( PL_get_chars(c, &s, CVT_ATOM|CVT_LIST|CVT_STRING) )
-  { Puts(s);
-  } else
-  { err:
-    return PL_error("put", 1, NULL, ERR_TYPE, ATOM_character, c);
-  }
+    return Puts(s);
 
-  succeed;
+  return PL_error("put", 1, NULL, ERR_TYPE, ATOM_character, c);
 }
 
 word
@@ -1319,11 +1359,13 @@ setUnifyStreamNo(term_t stream, int n)
       
 
 static const opt_spec open4_options[] = 
-{ { ATOM_type,       OPT_ATOM },
-  { ATOM_reposition, OPT_BOOL },
-  { ATOM_alias,	     OPT_ATOM },
-  { ATOM_eof_action, OPT_ATOM },
-  { NULL_ATOM,	     0 }
+{ { ATOM_type,		 OPT_ATOM },
+  { ATOM_reposition,     OPT_BOOL },
+  { ATOM_alias,	         OPT_ATOM },
+  { ATOM_eof_action,     OPT_ATOM },
+  { ATOM_close_on_abort, OPT_BOOL },
+  { ATOM_buffer,	 OPT_ATOM },
+  { NULL_ATOM,	         0 }
 };
 
 
@@ -1332,14 +1374,17 @@ pl_open4(term_t file, term_t mode,
 	 term_t stream, term_t options)
 { int m = -1;
   atom_t mname;
-  atom_t type       = ATOM_text;
-  bool reposition = FALSE;
-  atom_t alias	  = NULL_ATOM;
-  atom_t eof_action = ATOM_eof_code;
-  int flags = OPEN_OPEN;
+  atom_t type           = ATOM_text;
+  bool   reposition     = FALSE;
+  atom_t alias	        = NULL_ATOM;
+  atom_t eof_action     = ATOM_eof_code;
+  atom_t buffer         = ATOM_full;
+  bool   close_on_abort = TRUE;
+  int	 flags          = OPEN_OPEN;
 
   if ( !scan_options(options, 0, ATOM_stream_option, open4_options,
-		     &type, &reposition, &alias, &eof_action) )
+		     &type, &reposition, &alias, &eof_action,
+		     &close_on_abort, &buffer) )
     fail;
 
   if ( alias )
@@ -1368,13 +1413,16 @@ pl_open4(term_t file, term_t mode,
 
     if ( openStream(file, m, flags) )
     { if ( setUnifyStreamNo(stream, Input) )
-      { if ( eof_action != ATOM_eof_code )
-	{ IOSTREAM *s = fileTable[Input].stream;
-	  if ( eof_action == ATOM_reset )
+      { IOSTREAM *s = fileTable[Input].stream;
+	  
+	if ( eof_action != ATOM_eof_code )
+	{ if ( eof_action == ATOM_reset )
 	    s->flags |= SIO_NOFEOF;
 	  else if ( eof_action == ATOM_error )
 	    s->flags |= SIO_FEOF2ERR;
 	}
+	if ( !close_on_abort )
+	  s->flags |= SIO_NOCLOSE;
 	Input = in;
 	pushInputContext();
         succeed;
@@ -1390,7 +1438,19 @@ pl_open4(term_t file, term_t mode,
   { int out = Output;
     if ( openStream(file, m, flags) )
     { if ( setUnifyStreamNo(stream, Output) )
-      { Output = out;
+      { IOSTREAM *s = fileTable[Output].stream;
+
+	if ( !close_on_abort )
+	  s->flags |= SIO_NOCLOSE;
+	if ( buffer != ATOM_full )
+	{ s->flags &= ~SIO_FBUF;
+	  if ( buffer == ATOM_line )
+	    s->flags |= SIO_LBUF;
+	  if ( buffer == ATOM_false )
+	    s->flags |= SIO_NBUF;
+	}
+
+	Output = out;
         succeed;
       }
       closeStream(Output);
@@ -1531,21 +1591,25 @@ word
 pl_close(term_t stream)
 { int n;
   int isread;
+  int rval;
 
   if ( (n = streamNo(stream, F_ANY)) < 0 )
     fail;
   isread = (fileTable[n].status == F_READ);
 
-  TRY( closeStream(n) );
   if ( isread )
+  { rval = closeStream(n);
     popInputContext();
+  } else
+  { TRYPIPE(n, ATOM_write, rval = closeStream(n), rval = FALSE);
+  }
   
   if ( n == Output )
     Output = 1;
   if ( n == Input )
     Input = 0;
 
-  succeed;
+  return rval;
 }
 
 word
@@ -1593,7 +1657,7 @@ pl_flush_output(term_t stream)
 
   if ( (n = streamNo(stream, F_WRITE)) < 0 )
     fail;
-  Sflush(fileTable[n].stream);
+  TRYPIPE(n, ATOM_write, Sflush(fileTable[n].stream), fail);
 
   succeed;
 }
@@ -1663,6 +1727,41 @@ pl_stream_position(term_t stream, term_t old, term_t new)
 }
 
 
+word
+pl_seek(term_t stream, term_t offset, term_t method, term_t newloc)
+{ int n;
+  atom_t m;
+  int whence = -1;
+  long off, new;
+
+  if ( (n = streamNo(stream, F_ANY)) < 0 )
+    fail;
+  if ( !(PL_get_atom(method, &m)) )
+    goto badmethod;
+
+  if ( m == ATOM_bof )
+    whence = SIO_SEEK_SET;
+  else if ( m == ATOM_current )
+    whence = SIO_SEEK_CUR;
+  else if ( m == ATOM_eof )
+    whence = SIO_SEEK_END;
+  else
+  { badmethod:
+    return PL_error("seek", 4, NULL, ERR_DOMAIN, ATOM_seek_method, method);
+  }
+  
+  if ( !PL_get_long(offset, &off) )
+    return PL_error("seek", 4, NULL, ERR_DOMAIN, ATOM_integer, offset);
+
+  new = Sseek(fileTable[n].stream, off, whence);
+  if ( new == -1 )
+    return PL_error("seek", 4, OsError(), ERR_PERMISSION,
+		    ATOM_reposition, ATOM_stream, stream);
+
+  return PL_unify_integer(newloc, new);
+}
+
+
 word
 pl_set_input(term_t stream)
 { int n;
@@ -1770,14 +1869,17 @@ word
 pl_peek_byte2(term_t stream, term_t chr)
 { int n;
   IOSTREAM *s;
+  IOPOS pos;
   int c;
 
   if ( (n = streamNo(stream, F_READ)) < 0 ||
        !(s = fileTable[n].stream) )
     fail;
 
+  pos = s->posbuf;
   c = Sgetc(s);
   Sungetc(c, s);
+  s->posbuf = pos;
 
   return PL_unify_integer(chr, c);
 }
@@ -1786,13 +1888,16 @@ pl_peek_byte2(term_t stream, term_t chr)
 word
 pl_peek_byte1(term_t chr)
 { IOSTREAM *s;
+  IOPOS pos;
   int c;
 
   if ( !(s = fileTable[Input].stream) )
     fail;
 
+  pos = s->posbuf;
   c = Sgetc(s);
   Sungetc(c, s);
+  s->posbuf = pos;
 
   return PL_unify_integer(chr, c);
 }
@@ -1834,6 +1939,36 @@ PL_get_filename(term_t n, char *buf, unsigned int size)
 }
 
 
+int
+PL_get_stream_handle(term_t spec, IOSTREAM **stream)
+{ int n = streamNo(spec, F_ANY);
+
+  if ( n >= 0 )
+  { *stream = fileTable[n].stream;
+    succeed;
+  }
+
+  fail;
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+This should not exists (actually, the  integer stream handles should not
+exists). Used in PL_open_resource() to  get   rid  of the Prolog stream,
+while maintaining the IOSTREAM handle.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+void
+release_stream_handle(term_t spec)
+{ int n = streamNo(spec, F_ANY);
+  PlFile f = &fileTable[n];
+
+  f->stream = NULL_ATOM;
+  f->name   = NULL_ATOM;
+  f->status = F_CLOSED;
+}
+
+
 word
 pl_time_file(term_t name, term_t t)
 { char *fn;
@@ -1869,6 +2004,17 @@ pl_size_file(term_t name, term_t len)
 }
 
 
+word
+pl_size_stream(term_t stream, term_t len)
+{ int n;
+
+  if ( (n = streamNo(stream, F_ANY)) < 0 )
+    fail;
+
+  return PL_unify_integer(len, Ssize(fileTable[n].stream));
+}
+
+
 word
 pl_access_file(term_t name, term_t mode)
 { char *n;
@@ -2197,6 +2343,17 @@ pl_prolog_to_os_filename(term_t pl, term_t os)
 }
 
 
+foreign_t
+pl_mark_executable(term_t path)
+{ char name[MAXPATHLEN];
+
+  if ( !PL_get_filename(path, name, sizeof(name)) )
+    return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_source_sink, path);
+
+  return MarkExecutable(name);
+}
+
+
 #if defined(O_XOS) && defined(__WIN32__)
 word
 pl_make_fat_filemap(term_t dir)
