@@ -1,4 +1,4 @@
-/*  $Id: pl-flag.c,v 1.11 1997/12/03 08:52:00 jan Exp $
+/*  $Id: pl-flag.c,v 1.13 1998/07/06 10:29:46 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -70,10 +70,10 @@ pl_flag(term_t name, term_t old, term_t new)
       TRY(PL_unify_integer(old, f->value.i));
       break;
     case FLG_REAL:
-    {
+    { 
 #ifdef DOUBLE_ALIGNMENT
       double v;
-      memcpy(&v, &f->value.f, sizeof(double));
+      doublecpy(&v, &f->value.f);
       TRY(PL_unify_float(old, v));
 #else
       TRY(PL_unify_float(old, f->value.f));
@@ -99,7 +99,7 @@ pl_flag(term_t name, term_t old, term_t new)
     { 
       f->type = FLG_REAL;
 #ifdef DOUBLE_ALIGNMENT
-      memcpy(&f->value.f, &n.value.f, sizeof(double));
+      doublecpy(&f->value.f, &n.value.f);
 #else
       f->value.f = n.value.f;
 #endif
