@@ -1,4 +1,4 @@
-/*  $Id: pl-os.h,v 1.37 1997/10/13 10:08:46 jan Exp $
+/*  $Id: pl-os.h,v 1.40 1998/07/06 10:30:11 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -27,7 +27,6 @@
 
 #define allocHeap(n)	alloc_heap((alloc_t) (n))
 #define freeHeap(p, n)	free_heap((char *)(p), (alloc_t)(n))
-#define allocGlobal(n)	alloc_global((int) (n))
 
 extern Void Allocate(long);
 
@@ -35,9 +34,6 @@ extern Void Allocate(long);
 		*         MISCELLANEOUS         *
 		*********************************/
 
-extern char *Setenv(char *name, char *value);
-extern char *Unsetenv(char *name);
-
 extern char *OsError(void);
 extern bool initOs(void);
 extern volatile void Halt(int);
@@ -87,14 +83,14 @@ extern long Random(void);
 		 *******************************/
 
 #ifdef HAVE_SYSCONF
-#ifdef _SC_OPEN_MAX			/* getdtablesize() */
+#if defined(_SC_OPEN_MAX) && !defined(HAVE_GETPAGESIZE)
 #undef getdtablesize
 #define getdtablesize() sysconf(_SC_OPEN_MAX)
 #ifndef HAVE_GETDTABLESIZE
 #define HAVE_GETDTABLESIZE 1
 #endif
 #endif
-#ifdef _SC_PAGESIZE			/* getpagesize */
+#if defined(_SC_PAGESIZE) && !defined(HAVE_GETPAGESIZE)
 #undef getpagesize
 #define getpagesize() sysconf(_SC_PAGESIZE)
 #ifndef HAVE_GETPAGESIZE
@@ -123,7 +119,7 @@ extern int	getpagesize(void);
 		 *               TYPES		*
 		 *******************************/
 
-#if defined(__sun__) && defined(__svr4__)
+#if 0 && defined(__sun__) && defined(__svr4__)
 extern long random(void);
 extern int srandom(unsigned seed);
 #endif
