@@ -24,33 +24,11 @@
 
 forwards void profile(int);
 
-struct itimerval value, ovalue;		/* itmer controlling structures */
-
-#ifdef HAVE_SIGACTION
-static struct sigaction osigaction;
-#endif
-
-#ifndef SA_NOMASK
-#define SA_NOMASK 0
-#endif
-#ifndef SA_RESTART
-#define SA_RESTART 0
-#endif
+struct itimerval value, ovalue;		/* itimer controlling structures */
 
 static bool
 startProfiler(int how)
-{
-#ifdef HAVE_SIGACTION
-  struct sigaction action;
-
-  memset((char *) &action, 0, sizeof(action));
-  action.sa_handler  = profile;
-  action.sa_flags    = SA_NOMASK|SA_RESTART;
-
-  sigaction(SIGPROF, &action, &osigaction);
-#else
-  pl_signal(SIGPROF, profile);
-#endif
+{ set_sighandler(SIGPROF, profile);
 
   value.it_interval.tv_sec  = 0;
   value.it_interval.tv_usec = 1;
@@ -86,15 +64,7 @@ stopProfiler()
 
   stopItimer();
   LD->statistics.profiling = NO_PROFILING;
-#ifdef HAVE_SIGACTION
-  sigaction(SIGPROF, &osigaction, NULL);
-#else
-#ifdef _AIX
-  pl_signal(SIGPROF, SIG_IGN);
-#else
-  pl_signal(SIGPROF, SIG_DFL);
-#endif
-#endif
+  set_sighandler(SIGPROF, SIG_IGN);
 
   succeed;
 }
