@@ -1,4 +1,4 @@
-/*  $Id: pl-write.c,v 1.23 1997/12/03 08:52:05 jan Exp $
+/*  $Id: pl-write.c,v 1.27 1998/08/06 13:27:37 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -81,6 +81,26 @@ PutToken(const char *s)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+PutOpenBrace()/PutCloseBrace() are used to put  additional braces around
+a term to avoid an operator  precedence   problem.  If  the last emitted
+character  is  alphanumerical,  there  should  be  a  space  before  the
+openbrace to avoid interpretation as a term.   E.g. not (a,b) instead of
+not(a,b).  Reported by Stefan.Mueller@dfki.de.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static bool
+PutOpenBrace()
+{ return PutOpenToken('(') && Put('(');
+}
+
+
+static bool
+PutCloseBrace()
+{ return Put(')');
+}
+
+
 static bool
 writeAtom(atom_t a, bool quote)
 { char *s = stringAtom(a);
@@ -95,14 +115,59 @@ writeAtom(atom_t a, bool quote)
       case AT_QUOTE:
       case AT_FULLSTOP:
       default:
-      { char c;
+      { int c;
 
 	TRY(Put('\''));
 	while( (c = *s++) != EOS )
-	{ if (c == '\'')
-	  { TRY(Put(c)&&Put(c));
+	{ if ( trueFeature(CHARESCAPE_FEATURE) )
+	  { if ( c >= ' ' && c != 127 && c != '\'' )
+	    { TRY(Put(c));
+	    } else
+	    { char esc[4];
+
+	      esc[1] = EOS;
+
+	      switch(c)
+	      { case 7:
+		  esc[0] = 'a';
+		  break;
+		case '\b':
+		  esc[0] = 'b';
+		  break;
+		case '\t':
+		  esc[0] = 't';
+		  break;
+		case '\n':
+		  esc[0] = 'n';
+		  break;
+		case 11:
+		  esc[0] = 'v';
+		  break;
+		case '\r':
+		  esc[0] = 'r';
+		  break;
+		case '\f':
+		  esc[0] = 'f';
+		  break;
+		case '\\':
+		  esc[0] = '\\';
+		  break;
+		case '\'':
+		  esc[0] = '\'';
+		  break;
+		default:
+		  Ssprintf(esc, "%03o", c);
+	      }
+	      if ( !Put('\\') ||
+		   !Puts(esc) )
+		fail;
+	    }
 	  } else
-	  { TRY(Put(c));
+	  { if ( c == '\'' )
+	    { TRY(Put(c)&&Put(c));
+	    } else
+	    { TRY(Put(c));
+	    }
 	  }
 	}
 	return Put('\'');
@@ -135,14 +200,14 @@ writePrimitive(term_t t, bool quote)
   { long i;
 
     PL_get_long(t, &i);
-    PutOpenToken('0');			/* Any Alpha char will do */
+    PutOpenToken(i < 0 ? '-' : '0');	/* Any Alpha char will do */
     return Putf("%ld", i);
   }
 
   if ( PL_get_float(t, &f) )
   { char *s = NULL;
 
-    PutOpenToken('0');			/* Any Alpha char will do */
+    PutOpenToken(f < 0.0 ? '-' : '0');	/* Any Alpha char will do */
 
 #ifdef HUGE_VAL
     if ( f == HUGE_VAL )
@@ -266,9 +331,9 @@ writeTerm2(term_t t, int prec, bool style)
 
   if ( PL_get_atom(t, &a) )
   { if ( priorityOperator(a) > prec )
-    { if ( Put('(') &&
+    { if ( PutOpenBrace() &&
 	   writeAtom(a, quote) &&
-	   Put(')') )
+	   PutCloseBrace() )
 	succeed;
     } else
       return writeAtom(a, quote);
@@ -321,12 +386,12 @@ writeTerm2(term_t t, int prec, bool style)
   
 	  PL_get_arg(1, t, arg);
 	  if ( op_pri > prec )
-	  { TRY(Put('('));
+	  { TRY(PutOpenBrace());
 	  }
 	  TRY(writeAtom(functor, quote));
 	  TRY(writeTerm2(arg, op_type == OP_FX ? op_pri-1 : op_pri, style));
 	  if ( op_pri > prec )
-	  { TRY(Put(')'));
+	  { TRY(PutCloseBrace());
 	  }
 
 	  succeed;
@@ -338,11 +403,11 @@ writeTerm2(term_t t, int prec, bool style)
   
 	  PL_get_arg(1, t, arg);
 	  if ( op_pri > prec )
-	    TRY(Put('('));
+	    TRY(PutOpenBrace());
 	  TRY(writeTerm2(arg, op_type == OP_XF ? op_pri-1 : op_pri, style));
 	  TRY(writeAtom(functor, quote));
 	  if (op_pri > prec)
-	    TRY(Put(')'));
+	    TRY(PutCloseBrace());
   
 	  succeed;
 	}
@@ -377,7 +442,7 @@ writeTerm2(term_t t, int prec, bool style)
 	  PL_get_arg(2, t, r);
   
 	  if ( op_pri > prec )
-	    TRY(Put('('));
+	    TRY(PutOpenBrace());
 	  TRY(writeTerm2(l, 
 			op_type == OP_XFX || op_type == OP_XFY
 				? op_pri-1 : op_pri, 
@@ -390,7 +455,7 @@ writeTerm2(term_t t, int prec, bool style)
 				? op_pri-1 : op_pri, 
 			style));
 	  if ( op_pri > prec )
-	    TRY(Put(')'));
+	    TRY(PutCloseBrace());
 	  succeed;
 	}
       }
