@@ -1,4 +1,4 @@
-/*  $Id: mmap.c,v 1.9 1997/11/12 14:21:29 jan Exp $
+/*  $Id: mmap.c,v 1.10 1998/07/06 10:30:53 jan Exp $
 
     Designed and implemented by Jan Wielemaker
     E-mail: jan@swi.psy.uva.nl
@@ -184,6 +184,16 @@ getpagesize()
 #endif
 #ifdef HAVE_SYS_RESOURCE_H
 #include <sys/resource.h>
+#endif
+
+					/* __linux__ TASK_SIZE stuff */
+					/* contributed by Roman Hodek */
+#ifdef __linux__
+#include <asm/page.h>
+#include <asm/system.h>
+#include <asm/ptrace.h>
+#include <asm/segment.h>
+#include <asm/processor.h>
 #endif
 
 #ifdef RLIMIT_DATA
@@ -194,6 +204,14 @@ topOfHeap(ulong heap_base)
   if ( getrlimit(RLIMIT_DATA, &limit) == 0 )
   { ulong top = limit.rlim_cur + heap_base;
 
+#ifdef TASK_SIZE
+    if ( top < heap_base || top > TASK_SIZE )
+      return TASK_SIZE;
+#else
+    if ( top < heap_base )
+      return 0L;
+#endif
+
 #ifdef VERBOSE
     printf("Heap: %p ... %p\n", (void *)heap_base, (void *)top);
 #endif
